from threading import Thread


def async_run(f):
    """
    异步执行修饰器，通过如下方式让函数非阻塞异步执行

    @async_run
    def func():
        print("run async!")

    :param f:
    :return:
    """
    def wrapper(*args, **kwargs):
        t = Thread(target=f, args=args, kwargs=kwargs)
        t.setDaemon(True)
        t.start()

    return wrapper



import pqsdk.faxconstant
from pqsdk.enums.orderStatus import OrderStatus


class Constant(object):
    """
    常量实例，用于存储策略中用的的系统指定的常量
    """

    @property
    def OrderStatus(self):
        return OrderStatus

    @property
    def LATEST_PRICE(self):
        """
        # 最新价
        :return:
        """
        return faxconstant.LATEST_PRICE

    @property
    def FIX_PRICE(self):
        """
        # 指定价/限价
        :return:
        """
        return faxconstant.FIX_PRICE

    @property
    def MARKET_SH_CONVERT_5_CANCEL(self):
        """
        最优五档即时成交剩余撤销[上交所][股票]
        :return:
        """
        return faxconstant.MARKET_SH_CONVERT_5_CANCEL

    @property
    def MARKET_SH_CONVERT_5_LIMIT(self):
        """
        最优五档即时成交剩转限价[上交所][股票]
        :return:
        """
        return faxconstant.MARKET_SH_CONVERT_5_LIMIT

    @property
    def MARKET_PEER_PRICE_FIRST(self):
        """
        对手方最优价格委托[上交所[股票]][深交所[股票][期权]]
        :return:
        """
        return faxconstant.MARKET_PEER_PRICE_FIRST

    @property
    def MARKET_MINE_PRICE_FIRST(self):
        """
        本方最优价格委托[上交所[股票]][深交所[股票][期权]]
        :return:
        """
        return faxconstant.MARKET_MINE_PRICE_FIRST

    @property
    def MARKET_SZ_INSTBUSI_RESTCANCEL(self):
        """
        即时成交剩余撤销委托[深交所][股票][期权]
        :return:
        """
        return faxconstant.MARKET_SZ_INSTBUSI_RESTCANCEL

    @property
    def MARKET_SZ_CONVERT_5_CANCEL(self):
        """
        最优五档即时成交剩余撤销[深交所][股票][期权]
        :return:
        """
        return faxconstant.MARKET_SZ_CONVERT_5_CANCEL

    @property
    def MARKET_SZ_FULL_OR_CANCEL(self):
        """
        全额成交或撤销委托[深交所][股票][期权]
        :return:
        """
        return faxconstant.MARKET_SZ_FULL_OR_CANCEL



from six import with_metaclass
import abc
from typing import Callable
from .abstractPosition import AbstractPosition
from .abstractInstrument import AbstractInstrument
import datetime
import pandas as pd


class AbstractStrategyContext(with_metaclass(abc.ABCMeta)):
    """
    策略上下文的抽象接口类。
    """

    @abc.abstractmethod
    def plot_line(self, sec_code: str, name: str, value: float, desc=None):
        """
        画曲线
        :param sec_code:
        :param name:
        :param value:
        :param desc:
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def plot_bar(self, sec_code: str, name: str, value: float, desc=None):
        """
        画柱状图
        :param sec_code:
        :param name:
        :param value:
        :param desc:
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run_monthly(self, func: Callable, monthday: int, time_str: str) -> None:
        """
        按月运行

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param monthday: 每月的第几个交易日, day (1-31)
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run_weekly(self, func: Callable, weekday: int, time_str: str) -> None:
        """
        按周运行

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param weekday: 每周的第几个交易日, 1 = monday, ... 7 = sunday
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run_periodically(self,
                         func: Callable,
                         days: int,
                         time_str: str,
                         start_date: str = None,
                         end_date: str = None) -> None:
        """
        按调仓周期运行，如果遇到调仓日为非交易日，则顺延到下一个交易日。

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param days : 间隔天数执行，start_date为第一次执行日期
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；在实盘中才生效
        :param start_date: 开始日期，默认为回测开始日期
        :param end_date: 结束日期, 默认为回测结束日期
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run_daily(self, func: Callable, time_str: str) -> None:
        """
        每天内何时运行

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run_minutely(self, func: Callable, time_str: str, minutes: int = 1) -> None:
        """
        每天内何时开始，间隔n分钟执行

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；
        :param minutes: 间隔执行的分钟数
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run_secondly(self, func: Callable, time_str: str, seconds: int = 1) -> None:
        """
        每天内何时开始，间隔n秒执行

        :param func: 一个自定义的函数, 此函数必须接受context参数;例如自定义函数名market_open(context)
        :param time_str: 具体执行时间,一个字符串格式的时间, 24小时内的任意时间，例如"10:00", "01:00"；
        :param seconds: 间隔执行的秒数
        """
        raise NotImplementedError

    @abc.abstractmethod
    def cancel(self, order, force: bool = False):
        """
        取消订单委托
        :param force: 强制交易
        :param order: 回测中为order对象，实盘中为order_id
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def close(self, stock_code: str, force: bool = False):
        """
        以最新价平仓股票代码, 默认平仓数量为可用数量
        :param force: 强制交易
        :param stock_code: 需要平仓的股票代码
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_current_data(self):
        """
        获取当前单位时间（当天/当前分钟）的涨跌停价, 是否停牌，当天的开盘价等

        :return: 一个dict, 其中 key 是股票代码, value 是拥有如下属性的对象, 返回的结果只在当天有效:
                last_price : 最新价
                high_limit: 涨停价
                low_limit: 跌停价
                paused: 是否停止或者暂停了交易, 当停牌、未上市或者退市后返回 True
                is_st: 是否是 ST(包括ST, *ST)，是则返回 True，否则返回 False
                day_open: 当天开盘价
                name: 股票现在的名称, 可以用这个来判断股票当天是否是 ST, *ST, 是否快要退市
                industry_code: 股票现在所属行业代码
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_position(self, stock_code) -> AbstractPosition:
        """
        获取已经登录账号的股票持仓，返回个股持仓字典

        :param stock_code:
        :return: {'volume': 72000, 'can_use_volume': 0, 'open_price': 10.8, 'market_value': 1020240.0}
        """
        raise NotImplementedError

    @abc.abstractmethod
    def order_target_value(self,
                           stock_code: str,
                           target: float = 0.0,
                           price: float = None,
                           force: bool = False):
        """
        Place an order to rebalance a position to have final value of
        ``target``

        The current ``value`` is taken into account as the start point to
        achieve ``target``

          - If no ``target`` then close postion on data
          - If ``target`` > ``value`` then buy on data
          - If ``target`` < ``value`` then sell on data

        It returns either:

          - The generated order

          or

          - ``None`` if no order has been issued

        :param force: 强制交易
        :param stock_code: 目标持仓股票
        :param target: 目标持仓市值
        :param price:
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def order_target_percent(self,
                             stock_code: str,
                             target: float = 0.0,
                             force: bool = False):
        """
        基于order_target_value（），按照总资产的百分比委托下单

        :param force: 强制交易
        :param stock_code:目标持仓股票
        :param target: 目标持仓比例
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def order_target_volume(self,
                            stock_code: str,
                            target: float = 0.0,
                            force: bool = False):
        """
        基于order_target_value()，按照个股的数量目标数量下单。注意，可能存在账户资金不足的情况。

        :param force: 强制交易
        :param stock_code: 目标持仓股票
        :param target: 目标持仓数量
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def constant(self):
        """
        上下文的系统常量
        :return:
        """
        raise NotImplementedError

    # @property
    # @abc.abstractmethod
    # def planned_trade_dates(self) -> list:
    #     """
    #     获取调仓交易日列表。
    #     :return:
    #     """
    #     raise NotImplementedError

    @property
    @abc.abstractmethod
    def previous_trade_date(self) -> datetime.date:
        """
        当前bar对应的前一个交易日
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def parameters(self) -> dict:
        """
        执行策略的参数字典
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def stock_pool(self):
        """
        股票池
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def commission(self):
        """
        佣金费率
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def strategy_name(self):
        """
        策略名称
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def unit(self):
        """
        unit: 单位时间长度，支持1d、1m，默认为1d
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def dividend_type(self):
        """
        dividend_type: 复权选项(对股票/基金的价格字段、成交量字段及factor字段生效)
                'front'
                : 前复权, 默认是前复权
                none
                : 不复权, 返回实际价格
                'back'
                : 后复权
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def current_dt(self) -> datetime.datetime:
        """
        当前bar对应的日期/时间, format: %Y-%m-%d %H:%M:%S
        :return:
        """
        raise NotImplementedError

    # @property
    # @abc.abstractmethod
    # def next_trade_datetime(self):
    #     """
    #     当前bar对应的下一个日期/时间, format: %Y-%m-%d %H:%M:%S
    #     :return:
    #     """
    #     raise NotImplementedError

    @abc.abstractmethod
    def buy(self,
            stock_code: str,
            volume: int,
            price_type: int,
            price: float,
            force: bool = False):
        """
        证券买入委托
        Returns:
          - 委托的order_id
        """
        raise NotImplementedError

    @abc.abstractmethod
    def sell(self,
             stock_code: str,
             volume: int,
             price_type: int,
             price: float,
             force: bool = False):
        """
        证券卖出委托
        Returns:
          - 委托的order_id
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def portfolio(self):
        """
        策略投资组合，可通过该对象获取当前策略账户、持仓等信息
        """
        raise NotImplementedError

    @abc.abstractmethod
    def inout_cash(self, cash: float):
        """
        投资组合转入或转出资金，当日的出入金从当日开始记入成本，用于计算收益，即当日结束计算收益时的本金是包含当日出入金金额的
        :param cash: 可正可负，正为入金，负为出金。
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def run_info(self):
        """
        策略运行信息，包括回测过程中的所有参数。
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_security_info(self, sec_code: str) -> AbstractInstrument:
        """
        获取股票/基金/指数/期货的信息

        :param sec_code: 证券代码
        :return: 一个dict, 有如下属性:
                display_name: 中文名称
                name: 缩写简称
                start_date: 上市日期
                end_date: 退市日期（股票是最后一个交易日，不同于摘牌日期）， [datetime.date] 类型, 如果没有退市则为2200-01-01
                type:   'index'		#指数
                        'stock'		#股票
                        'fund'		#基金
                        'etf'		#ETF
        """
        raise NotImplementedError

    @abc.abstractmethod
    def query_new_purchase_limit(self):
        """
        查询新股申购额度
        :return:
            dict 新股申购额度数据集
            { type1: number1, type2: number2, ... }
            type - str 品种类型
            KCB - 科创板，SH - 上海，SZ - 深圳
            number - int 可申购股数
        """
        raise NotImplementedError

    @abc.abstractmethod
    def query_ipo_data(self):
        """
        查询当日新股新债信息
        :return:
            dict 新股新债信息数据集
            { stock1: info1, stock2: info2, ... }
                stock - str 品种代码，例如 '301208.SZ'
                info - dict 新股信息
                name - str 品种名称
                type - str 品种类型
                STOCK - 股票，BOND - 债券
                minPurchaseNum / maxPurchaseNum - int 最小 / 最大申购额度
                单位为股（股票）/ 张（债券）
                purchaseDate - str 申购日期
                issuePrice - float 发行价
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_factor(self,
                   sec_code_list: list = None,
                   factor_list: list = ...,
                   trade_date: str = None,
                   start_date: str = None,
                   end_date: str = None,
                   stock_pool: list = None,
                   expect_df: bool = True) -> pd.DataFrame:
        """
        TODO 返回给定的 sec_code 当日的因子，包括财务因子、alpha101 因子、技术指标 等

        :param trade_date: 指定交易日日期，默认为该日的前一个交易日
        :param sec_code_list:  合约代码
        :param factor_list: 因子名称
        :param start_date: 开始日期
        :param end_date:  结束日期
        :param stock_pool: 股票池，可选定指数的成分股，默认为 None，全市场
        :param expect_df: 默认返回 pandas dataframe
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def history(self, count, unit='1d', field='avg', security_list=None, expect_df=True, skip_paused=False,
                dividend_type='pre'):
        """
        获取历史数据，可查询多个标的单个数据字段，返回数据格式为 DataFrame 或 Dict(字典)

        :param count: 数量, 返回的结果集的行数
        :param unit: 单位时间长度, 几天或者几分钟, 现在支持'1d','1m'
        :param field: 要获取的数据字段
        :param security_list: 要获取数据的股票列表
        :param expect_df: expect_df=True: [pandas.DataFrame]对象, 行索引是[datetime.datetime]对象, 列索引是股票代号
                   expect_df=False: dict, key是股票代码, value是一个numpy数组[numpy.ndarray]
        :param skip_paused: 是否跳过不交易日期(包括停牌, 未上市或者退市后的日期)
        :param dividend_type: 复权选项(对股票/基金的价格字段、成交量字段及factor字段生效)
                    'front'
                    : 前复权, 默认是前复权
                    none
                    : 不复权, 返回实际价格
                    'back'
                    : 后复权
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def attribute_history(self,
                          security: str,
                          count: int,
                          unit: str = '1d',
                          fields: list = None,
                          skip_paused: bool = True,
                          expect_df: bool = True,
                          dividend_type: str = 'front'):
        """
        获取历史数据，可查询单个标的多个数据字段，返回数据格式为 DataFrame 或 Dict(字典)

        :param security: 股票代码
        :param count: 数量, 返回的结果集的行数
        :param unit: 单位时间长度, 1d, 1m
        :param fields: 股票属性的list, 支持：['open', ' close', 'low', 'high', 'volume', 'money', 'factor',
                                           'high_limit',' low_limit', 'avg', ' pre_close', 'paused']
        :param skip_paused: 是否跳过不交易日期(包括停牌, 未上市或者退市后的日期).
        :param expect_df: 若是True, 返回[pandas.DataFrame], 否则返回一个dict, 具体请看下面的返回值介绍. 默认是True.
        :param dividend_type: 复权选项(对股票/基金的价格字段、成交量字段及factor字段生效)
                    'front'
                    : 前复权, 默认是前复权
                    none
                    : 不复权, 返回实际价格
                    'back'
                    : 后复权
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def synthesize_portfolio(self,
                             count: int = None,
                             stock_pool: list = None,
                             start_date: str = None,
                             end_date: str = None,
                             freq: int = 5,
                             trade_date_lst: list = None,
                             sec_code: str = None,
                             sec_code_list: list = None,
                             factor_list=None,
                             expect_df: bool = False,
                             exact_match: bool = False,
                             stock_size: int = 10):
        """
        获取参数组合（stock_pool, start_date, end_date, freq）最优的合成因子数据所产生的股票组合。

        :param count: 获取最新的n期调仓日数据
        :param stock_pool: 股票池
        :param start_date: 因子取数的开始日期，也是合成因子回测的开始日期
        :param end_date: 因子取数的结束日期，也是合成因子回测的结束日期
        :param freq: 调仓周期，单位为天，默认值为5
        :param trade_date_lst: 限定调仓日列表
        :param sec_code: 股票
        :param sec_code_list: 股票代码列表
        :param factor_list: 合成因子列表，默认返回完整的合成因子列表
        :param expect_df: 若是True, 返回[pandas.DataFrame], 否则返回一个dict, 具体请看下面的返回值介绍. 默认是True.
        :param exact_match: 是否精确匹配（stock_pool, start_date, end_date, freq），或者近似匹配
        :param stock_size: 股票组合的股票个数
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_blacklist(self) -> list:
        """
        获取租户下定义的股票列表黑名单，黑名单的股票不建议持仓

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_last_ticks(self, sec_code_lst: list, index: bool = False, fields: list = None):
        """
        获取当前最新的Tick数据。
        注意：不是历史某个回测时间点的Tick数据。

        :param sec_code_lst:
        :param fields: 返回的字段列表
        :param index: 是否设置sec_code为Index
        :return:
                [{
                'sec_code', 股票代码
                ’datetime'  tick时间字符串： %Y-%m-%d %H:%M:%S
                'open', 当日开盘价
                'high', 当日收盘价
                'low', 当日最低价
                'last_price', 当日最新价格
                'pre_close', 昨日收盘价
                'volume', 当日成交量，单位：手
                'amount', 当日成交额
                }]
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_plate_data(self, trade_date: str = None, end_time: str = "15:00", limit: int = 3) -> pd.DataFrame:
        """
        获取板块列表

        :param trade_date: 指定交易日的板块列表
        :param end_time: 指定板块排名的时间节点，从09:25开始到15:30结束，每个间隔5分钟
        :param limit: 按照板块强度倒序排名的板块数量
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_plate_members_lst(self, plate_code_list: list, trade_date: str = None):
        """
        获取板块的成分股
        :param plate_code_list:
        :param trade_date:
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_stock_name(self, sec_code: str, trade_date: str = None):
        """
        获取股票代码的最新名称

        :param sec_code:
        :param trade_date:
            None：默认为当天，如果当天不是交易日，则获取上一个交易日。
            'YYYY-mm-dd'：具体日期的最新股票名称
        :return: 股票名称
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_stock_price(self,
                        sec_code: str,
                        trade_date: str = None,
                        end_time: str = None,
                        dividend_type: str = "front"):
        """
        获取最新的股票价格

        :param dividend_type: 复权选项(对股票/基金的价格字段、成交量字段及factor字段生效)
                    'front'
                    : 前复权, 默认是前复权
                    none
                    : 不复权, 返回实际价格
                    'back'
                    : 后复权
        :param trade_date: 指定交易日, 格式: 2024-07-15
        :param end_time: 如果trade_date不是系统日期当日，可以指定历史数据的具体分钟，格式：09:30
        :param sec_code:
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_price_limit(self,
                        sec_code: str,
                        trade_date: str = None,
                        dividend_type: str = "front"):
        """
        股票代码涨跌停价格
        假设：
        - 600xxx, 601xxx, 603xxx：上海主板，涨跌停限制10%
        - 000xxx：深圳主板，涨跌停限制10%
        - 002xxx：深圳中小板，涨跌停限制10%
        - 300xxx：深圳创业板，涨跌停限制20%
        - 688xxx：上海科创板，涨跌停限制20%

        :param dividend_type: 复权选项(对股票/基金的价格字段、成交量字段及factor字段生效)
                    'front'
                    : 前复权, 默认是前复权
                    none
                    : 不复权, 返回实际价格
                    'back'
                    : 后复权
        :param sec_code:
        :param trade_date: 指定日期的涨跌停价格
        :return: (昨日收盘价格, 今日涨停价, 今日跌停价)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_auction_data(self,
                         security: str,
                         count: int = None,
                         unit: str = '1m',
                         trade_date: str = None,
                         start_datetime: str = None,
                         end_datetime: str = None,
                         fields=None):
        """
        获取股票的集合竞价数据。

        :param security: 股票代码
        :param count: 返回行数
        :param unit: 支持1m, 5m
        :param trade_date: 交易日
        :param start_datetime: 开始时间
        :param end_datetime: 结束时间
        :param fields: 获取字段，范围：['open', 'high', 'low', 'close', 'pre_close']
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_orders(self, trade_date: str = None, is_buy: bool = None):
        """
        获取交易日的委托单，返回DataFrame

        :param trade_date:
        :param is_buy:
        :return: DataFrame,
            字段：['order_code', 委托编号
                'sec_code',  股票代码
                'order_time', 委托时间
                'is_buy', 委托买卖方向：买单=1， 卖单=0
                'order_volume',  委托数量
                'order_price']  委托价格
        """
        raise NotImplementedError




# coding=utf-8


# --------------------------------------
# 常量定义模块
# 命名规范： 模块_功能_描述 = 值
# --------------------------------------

"""
市场类型
"""
# 上海市场
MARKET_SH = 0
# 深圳市场
MARKET_SZ = 1

"""
报价类型
"""
# 最新价
LATEST_PRICE = 5
# 指定价/限价
FIX_PRICE = 11
# 最优五档即时成交剩余撤销[上交所][股票]
MARKET_SH_CONVERT_5_CANCEL = 42
# 最优五档即时成交剩转限价[上交所][股票]
MARKET_SH_CONVERT_5_LIMIT = 43
# 对手方最优价格委托[上交所[股票]][深交所[股票][期权]]
MARKET_PEER_PRICE_FIRST = 44
# 本方最优价格委托[上交所[股票]][深交所[股票][期权]]
MARKET_MINE_PRICE_FIRST = 45
# 即时成交剩余撤销委托[深交所][股票][期权]
MARKET_SZ_INSTBUSI_RESTCANCEL = 46
# 最优五档即时成交剩余撤销[深交所][股票][期权]
MARKET_SZ_CONVERT_5_CANCEL = 47
# 全额成交或撤销委托[深交所][股票][期权]
MARKET_SZ_FULL_OR_CANCEL = 48

price_type = {
    LATEST_PRICE: "最新价",
    FIX_PRICE: "限价",
    MARKET_SH_CONVERT_5_CANCEL: "最优五档即时成交剩余撤销[上交所][股票]",
    MARKET_SH_CONVERT_5_LIMIT: "最优五档即时成交剩转限价[上交所][股票]",
    MARKET_PEER_PRICE_FIRST: "对手方最优价格委托[上交所[股票]][深交所[股票][期权]]",
    MARKET_MINE_PRICE_FIRST: "本方最优价格委托[上交所[股票]][深交所[股票][期权]]",
    MARKET_SZ_INSTBUSI_RESTCANCEL: "即时成交剩余撤销委托[深交所][股票][期权]",
    MARKET_SZ_CONVERT_5_CANCEL: "最优五档即时成交剩余撤销[深交所][股票][期权]",
    MARKET_SZ_FULL_OR_CANCEL: "全额成交或撤销委托[深交所][股票][期权]",
}

"""
请求结果错误码
"""
RESULT_OK_CODE_200 = 200
# 400报错
# 1、有可能是前端请求时参数拼接有问题
# 2、有可能时前端传给后端的参数类型与后端接收的参数类型不匹配，比如前端传String，后端用Integer接收,
# 3、有可能是后端的实体类没有配置无参构造方法
RESULT_ERROR_CODE_400 = 400

# 403错误
# 服务端理解你的请求，但是没有权限访问
RESULT_ERROR_CODE_403 = 403
# 404错误
# 找不到网页地址，即网页的地址不正确，还有一种可能就是后端配置不当，导致请求被拦截
RESULT_ERROR_CODE_404 = 404
# 500错误
# 明显的后端错误，可能需要前后端配合来修复
RESULT_ERROR_CODE_500 = 500
# 502错误
# 网关错误，服务器作为网关或代理，无法从上游服务器收到正确的响应
RESULT_ERROR_CODE_502 = 502

"""
请求结果状态
"""
RESULT_STATUS_OK = 'OK'
RESULT_STATUS_ERROR = 'ERROR'

"""
委托状态
"""
# 未报
ORDER_UNREPORTED: int = 48
# 待报
ORDER_WAIT_REPORTING: int = 49
# 已报
ORDER_REPORTED: int = 50
# 已报待撤
ORDER_REPORTED_CANCEL: int = 51
# 部成待撤
ORDER_PARTSUCC_CANCEL: int = 52
# 部撤
ORDER_PART_CANCEL: int = 53
# 已撤
ORDER_CANCELED: int = 54
# 部成
ORDER_PART_SUCC: int = 55
# 已成
ORDER_SUCCEEDED: int = 56
# 废单
ORDER_JUNK: int = 57
# 未知
ORDER_UNKNOWN: int = 255
# 异常
ORDER_ERROR: int = 244

order_status = {
    ORDER_UNREPORTED: '未报',
    ORDER_WAIT_REPORTING: '待报',
    ORDER_REPORTED: '已报',
    ORDER_REPORTED_CANCEL: '已报待撤',
    ORDER_PARTSUCC_CANCEL: '部成待撤',
    ORDER_PART_CANCEL: '部撤',
    ORDER_CANCELED: '已撤',
    ORDER_PART_SUCC: '部成',
    ORDER_SUCCEEDED: '已成',
    ORDER_JUNK: '废单',
    ORDER_UNKNOWN: '未知',
    ORDER_ERROR: '异常'
}

"""
账号类型
"""
# 期货
FUTURE_ACCOUNT = 1
# 股票
SECURITY_ACCOUNT = 2
# 信用
CREDIT_ACCOUNT = 3
# 期货期权
FUTURE_OPTION_ACCOUNT = 5
# 股票期权
STOCK_OPTION_ACCOUNT = 6
# 沪港通
HUGANGTONG_ACCOUNT = 7
# 深港通
SHENGANGTONG_ACCOUNT = 11

account_type = {
    FUTURE_ACCOUNT: '期货',
    SECURITY_ACCOUNT: '股票',
    CREDIT_ACCOUNT: '信用',
    FUTURE_OPTION_ACCOUNT: '期货期权',
    STOCK_OPTION_ACCOUNT: '股票期权',
    HUGANGTONG_ACCOUNT: '沪港通',
    SHENGANGTONG_ACCOUNT: '深港通'
}

"""
账号状态
"""
"""
账号状态
"""
# 无效
ACCOUNT_STATUS_INVALID = -1
# 正常
ACCOUNT_STATUS_OK = 0
# 连接中
ACCOUNT_STATUS_WAITING_LOGIN = 1
# 登陆中
ACCOUNT_STATUSING = 2
# 失败
ACCOUNT_STATUS_FAIL = 3
# 初始化中
ACCOUNT_STATUS_INITING = 4
# 数据刷新校正中
ACCOUNT_STATUS_CORRECTING = 5
# 收盘后
ACCOUNT_STATUS_CLOSED = 6
# 穿透副链接断开
ACCOUNT_STATUS_ASSIS_FAIL = 7
# 系统停用（总线使用-密码错误超限）
ACCOUNT_STATUS_DISABLEBYSYS = 8
# 用户停用（总线使用）
ACCOUNT_STATUS_DISABLEBYUSER = 9

account_status = {
    ACCOUNT_STATUS_INVALID: '无效',
    ACCOUNT_STATUS_OK: '正常',
    ACCOUNT_STATUS_WAITING_LOGIN: '连接中',
    ACCOUNT_STATUSING: '登陆中',
    ACCOUNT_STATUS_FAIL: '失败',
    ACCOUNT_STATUS_INITING: '初始化中',
    ACCOUNT_STATUS_CORRECTING: '数据刷新校正中',
    ACCOUNT_STATUS_CLOSED: '收盘后',
    ACCOUNT_STATUS_ASSIS_FAIL: '穿透副链接断开',
    ACCOUNT_STATUS_DISABLEBYSYS: '系统停用（总线使用-密码错误超限）',
    ACCOUNT_STATUS_DISABLEBYUSER: '用户停用（总线使用）'
}

"""
委托类型
"""
"""
ORDER TYPE
"""
STOCK_BUY = 23
STOCK_SELL = 24
CREDIT_BUY = 23  # 担保品买入
CREDIT_SELL = 24  # 担保品卖出
CREDIT_FIN_BUY = 27  # 融资买入
CREDIT_SLO_SELL = 28  # 融券卖出
CREDIT_BUY_SECU_REPAY = 29  # 买券还券
CREDIT_DIRECT_SECU_REPAY = 30  # 直接还券
CREDIT_SELL_SECU_REPAY = 31  # 卖券还款
CREDIT_DIRECT_CASH_REPAY = 32  # 直接还款
CREDIT_FIN_BUY_SPECIAL = 40  # 专项融资买入
CREDIT_SLO_SELL_SPECIAL = 41  # 专项融券卖出
CREDIT_BUY_SECU_REPAY_SPECIAL = 42  # 专项买券还券
CREDIT_DIRECT_SECU_REPAY_SPECIAL = 43  # 专项直接还券
CREDIT_SELL_SECU_REPAY_SPECIAL = 44  # 专项卖券还款
CREDIT_DIRECT_CASH_REPAY_SPECIAL = 45  # 专项直接还款

order_type = {
    STOCK_BUY: '买入',
    STOCK_SELL: '卖出',
    # CREDIT_BUY: '担保品买入',
    # CREDIT_SELL: '担保品卖出',
    # CREDIT_FIN_BUY: '融资买入',
    # CREDIT_SLO_SELL: '融券卖出',
    # CREDIT_BUY_SECU_REPAY: '买券还券',
    # CREDIT_DIRECT_SECU_REPAY: '直接还券',
    # CREDIT_SELL_SECU_REPAY: '卖券还款',
    # CREDIT_DIRECT_CASH_REPAY: '直接还款',
    # CREDIT_FIN_BUY_SPECIAL: '专项融资买入',
    # CREDIT_SLO_SELL_SPECIAL: '专项融券卖出',
    # CREDIT_BUY_SECU_REPAY_SPECIAL: '专项买券还券',
    # CREDIT_DIRECT_SECU_REPAY_SPECIAL: '专项直接还券',
    # CREDIT_SELL_SECU_REPAY_SPECIAL: '专项卖券还款',
    # CREDIT_DIRECT_CASH_REPAY_SPECIAL: '专项直接还款',
}

"""
Celery Task Status
"""
CELERY_TASK_STATUS_PROCESSING = 'processing'
CELERY_TASK_STATUS_FINISHED = 'finished'
CELERY_TASK_STATUS_FAILED = 'failed'
CELERY_TASK_STATUS_TERMINATED = 'terminated'

celery_task_status = {
    CELERY_TASK_STATUS_PROCESSING: '运行中',
    CELERY_TASK_STATUS_FINISHED: '已完成',
    CELERY_TASK_STATUS_FAILED: '已失败',
    CELERY_TASK_STATUS_TERMINATED: '被中止'
}

"""
Service Key
"""
SERVICE_KEY_TICK_SUBSCRIBE = "SYS:SERVICE:TICK:SUBSCRIBE"  # 订阅Tick
SERVICE_KEY_TICK_PROCESS = "SYS:SERVICE:TICK:PROCESS"  # 处理Tick
SERVICE_KEY_CALLBACK_PROCESS = "SYS:SERVICE:CALLBACK:PROCESS"  # 处理交易


"""
Service Status
"""
SERVICE_STATUS_STARTUP = 1  # 已启动
SERVICE_STATUS_CLOSED = 2  # 已关闭
SERVICE_STATUS_ABNORMAL = 3  # 有异常

service_status = {
    SERVICE_STATUS_STARTUP: "已启动",
    SERVICE_STATUS_CLOSED: "已关闭",
    SERVICE_STATUS_ABNORMAL: "有异常"
}

"""
委托队列状态
"""
QUEUED_ORDER_STATUS_PENDING = 10
QUEUED_ORDER_STATUS_REPORTED = 11
QUEUED_ORDER_STATUS_CANCELED = 12
QUEUED_ORDER_STATUS_EXPIRED = 13

queued_order = {
    QUEUED_ORDER_STATUS_PENDING: "待处理",
    QUEUED_ORDER_STATUS_REPORTED: "已委托",
    QUEUED_ORDER_STATUS_CANCELED: "已取消",
    QUEUED_ORDER_STATUS_EXPIRED: "已过期"
}

"""
交易终端代理角色
"""
BROKER_ROLE_TYPE_TRADER = 1  # 交易终端
BROKER_ROLE_TYPE_TICKER = 2  # 数据终端

broker_role_type = {
    BROKER_ROLE_TYPE_TRADER: "交易终端",
    BROKER_ROLE_TYPE_TICKER: "数据终端"
}


"""
默认的交易策略
"""
DEFAULT_STRATEGY_ID = -1

default_strategy = {
    DEFAULT_STRATEGY_ID: "默认交易策略"
}



# -*- coding: utf-8 -*-

__version__ = "0.0.39"


_VersionInfo = __import__("collections").namedtuple(
    "version_info", ["major", "minor", "micro"]
)
version_info = ([int(v) for v in __version__.split('.')[:3]] + [0] * 3)[:3]
version_info = _VersionInfo(*version_info)

version_str = "PinkQuant SDK version {}, python version {}".format(
    __version__,
    ".".join(str(item) for item in __import__("sys").version_info)
)



from abc import ABC, abstractmethod
import pandas as pd
import numpy as np
from pqsdk.api import get_history


class AbstractAnalyzer(ABC):
    """
    分析器抽象函数
    """
    context = None

    def set_context(self, context):
        """
        设置回测的上下文对象
        :param context:
        :return:
        """
        self.context = context
        return self

    @abstractmethod
    def start(self):
        """
        在回测开始之前调用,对应第0根bar
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def next(self):
        """
        策略正常运行阶段, 每个Bar执行一次
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def stop(self):
        """
        策略结束时执行
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def get_analysis(self):
        """
        获取分析器结果
        :return:
        """
        raise NotImplementedError


class TimeReturn(AbstractAnalyzer):
    """
    获取Portfolio的收益情况
    """

    def __init__(self):
        # datetimes
        self.datetimes = []
        # 持仓成本
        self.cost = []
        # 持仓市值
        self.market_values = []
        # 盈亏：profit and loss
        self.pnls = []
        # 本金
        self.principal = []
        # 现金
        self.cash = []
        # 总资产
        self.assets = []
        # 仓位
        self.position_size = []
        # 净值
        self.navs = []
        # 收益率
        self.strat_returns = []

        # benchmark returns
        self.bchmk_returns = []

    def start(self):
        pass

    def get_metrics(self):
        """
        根据投资的的委托数据，出入金数据和最新的价格数据，计算投资组合的重要Metrics，比如：
            1. cost：持仓交易成本，截止当前的持仓成本余额
            2. market_value：持仓市值，截止当前的持仓市值，由最新的持仓标的价格计算所得。如果投资组合已经被全部平仓，则持仓市值为0，此时持仓成本为正数为亏损，持仓成本为负数为盈利。
            3. pnl：盈亏(profit and loss)，等于持仓市值 - 持仓成本 = market_value - cost。
            4. principal：本金，截止目前为止的投资组合出入金汇总。
            5. cash：现金，即截止今日剩余现金，等于本金 - 成本 = principal - cost
            6. asset：资产，等于现金 + 持仓市值 = cash + market_value
            7. position_size：仓位，等于持仓市值/资产 = market_value / asset
            8. nav：净值(Net Asset Value)，等于资产/本金 = asset / principal
            9. 累计收益率：Cumulative Return = NAV - 1
        :return:
        """
        # 获取委托明细数据
        df = pd.DataFrame(self.context.orders)
        if df.empty:
            metrics_dict = {'cost': 0, 'market_value': 0, 'pnl': 0}
        else:
            df['cost'] = df['volume'] * df['price']
            df = df[["sec_code", "trade_date", "is_buy", "volume", "cost"]]

            # volume, amount: 买单为正数，卖单为负数
            df['volume'] = np.where(df['is_buy'], df['volume'], df['volume'] * -1)
            df['cost'] = np.where(df['is_buy'], df['cost'], df['cost'] * -1)

            # 计算每个股票的目前持仓sum(volume)、成本sum(cost)
            stock_df = df.groupby(['sec_code'])[['volume', 'cost']].sum()
            stock_df = stock_df.reset_index()

            # 获取目前持仓的最新股票价格
            sec_code_lst = stock_df['sec_code'].unique().tolist()
            if self.context.unit == '1d':
                price_df = get_history(start_date=self.context.current_dt.strftime('%Y-%m-%d'),
                                       end_date=self.context.current_dt.strftime('%Y-%m-%d'),
                                       unit=self.context.unit,
                                       field='close',
                                       security_list=sec_code_lst,
                                       dividend_type=self.context.dividend_type,
                                       expect_df=True)
            else:  # unit in ['1m', '5m']
                price_df = get_history(count=1,  # 防止数据库中的准确的分钟级别数据丢失的时候，获取前一分钟
                                       start_date=self.context.current_dt.strftime('%Y-%m-%d'),
                                       end_date=self.context.current_dt.strftime('%Y-%m-%d'),
                                       end_datetime=self.context.current_dt.strftime('%Y-%m-%d %H:%M:%S'),
                                       unit=self.context.unit,
                                       field='close',
                                       security_list=sec_code_lst,
                                       dividend_type=self.context.dividend_type,
                                       expect_df=True)
            price_df = price_df.T.reset_index()
            price_df.columns = ["sec_code", "last_price"]
            stock_df = pd.merge(stock_df, price_df, on='sec_code')

            # 计算最新的持仓市值，盈亏
            stock_df['market_value'] = stock_df['volume'] * stock_df['last_price']
            stock_df['pnl'] = stock_df['market_value'] - stock_df['cost']

            # 统计投资组合基本的Metrics
            metrics_dict = stock_df[['cost', 'market_value', 'pnl']].sum().to_dict()

        # 本金: 出入金汇总
        cash_df = pd.DataFrame(self.context.inout_cash_his)
        metrics_dict['principal'] = cash_df['cash'].sum()

        # 截止今日剩余现金: 本金汇总 - 成本汇总
        metrics_dict['cash'] = metrics_dict['principal'] - metrics_dict['cost']
        # 资产：
        metrics_dict['asset'] = metrics_dict['cash'] + metrics_dict['market_value']
        # 仓位
        metrics_dict['position_size'] = metrics_dict['market_value'] / metrics_dict['asset']
        # 净值: Net Asset Value
        metrics_dict['nav'] = metrics_dict['asset'] / metrics_dict['principal']

        return metrics_dict

    def next(self):
        if self.context.unit in ['1m', '5m']:
            self.datetimes.append(self.context.current_dt.strftime('%Y-%m-%d %H:%M:%S'))
        else:
            self.datetimes.append(self.context.current_dt.strftime('%Y-%m-%d'))

        # 获取Metrics
        metrics_dict = self.get_metrics()

        # 根据净值的变化计算收益率。如果算累计收益率：Cumulative Return = NAV - 1
        if len(self.navs) > 0:
            start_return = metrics_dict['nav'] / self.navs[-1] - 1

        else:
            start_return = 0.0

        self.cost.append(metrics_dict['cost'])
        self.market_values.append(metrics_dict['market_value'])
        self.pnls.append(metrics_dict['pnl'])
        self.principal.append(metrics_dict['principal'])
        self.cash.append(metrics_dict['cash'])
        self.assets.append(metrics_dict['asset'])
        self.position_size.append(metrics_dict['position_size'])
        self.navs.append(metrics_dict['nav'])
        self.strat_returns.append(start_return)
        self.bchmk_returns.append(self.context.benchmark_value)

    def stop(self):
        pass

    def get_analysis(self):
        return {"cost": pd.Series(self.cost, index=self.datetimes, name='returns'),
                "market_value": pd.Series(self.market_values, self.datetimes, name='total_values'),
                "pnl": pd.Series(self.pnls, self.datetimes, name='market_values'),
                "principal": pd.Series(self.principal, self.datetimes, name='cash'),
                "cash": pd.Series(self.cash, self.datetimes, name='pnls'),
                "asset": pd.Series(self.assets, self.datetimes, name='bchmk_returns'),
                "position_size": pd.Series(self.position_size, self.datetimes, name='bchmk_returns'),
                "nav": pd.Series(self.navs, self.datetimes, name='bchmk_returns'),
                "strat_return": pd.Series(self.strat_returns, self.datetimes, name='bchmk_returns'),
                "bchmk_return": pd.Series(self.bchmk_returns, self.datetimes, name='bchmk_returns'),
                }



from pqsdk.interface import AbstractInstrument, AbstractRunInfo
import datetime
from pqsdk import log
from pqsdk.api import get_history, get_factor, get_stock_info, get_suspend_info, get_stock_name_his


class Instrument(AbstractInstrument):

    def __init__(self, stock_code: str, run_info: AbstractRunInfo, context):
        self.stock_code = stock_code
        self.context = context
        self.run_info = run_info

        # 获取股票基础信息
        bsc_info_df = get_stock_info(fields=['list_date', 'fullname', 'name', 'industry'],
                                     stock_lst=[stock_code])
        bsc_info_df = bsc_info_df.set_index(['sec_code'])
        self.bsc_info_dict = bsc_info_df.T.to_dict()

    @property
    def last_price(self) -> float:
        # 获取当前收盘价
        # TODO 没有缓存作用，每次访问查询数据库
        # trade_date = self.context.current_dt.strftime('%Y-%m-%d')
        if self.context.unit in ['1d']:
            trade_date = self.context.current_dt.strftime('%Y-%m-%d')
            trade_datetime = None
        else:
            trade_date = None
            trade_datetime = self.context.current_dt.strftime('%Y-%m-%d %H:%M:%S')

        price_df = get_history(count=1,
                               end_date=trade_date,
                               end_datetime=trade_datetime,
                               unit=self.context.unit,
                               field='close',
                               security_list=[self.stock_code],
                               dividend_type=self.context.dividend_type,
                               expect_df=True)
        if price_df.empty:
            log.warning(f"股票停牌，无法获取收盘价: stock_code={self.stock_code}, trade_date={trade_date}")
            price = 0.0
        else:
            price = price_df.iloc[0, 0]

        return price

    @property
    def high_limit(self) -> float:
        # 获取当前交易时间
        # TODO 没有缓存作用，每次访问查询数据库
        trade_date = self.context.current_dt.strftime('%Y-%m-%d')
        df = get_factor(self.stock_code, 'up_limit', trade_date=trade_date, dividend_type=self.context.dividend_type)
        if not df.empty:
            price = df.iloc[0, 0]
        else:
            log.warning(f"无法获取涨停价：sec_code={self.stock_code}, trade_date={trade_date}")
            price = 0.0
        return price

        # # 获取基础行情数据
        # fields = ['up_limit', 'down_limit']
        # # stk_d = crud.get_mkt_stk_d(trade_date=trade_date, fields=fields, stock_lst=[self.stock_code])
        # stk_d = crud.get_mkt_stk_d(trade_date=trade_date, fields=fields, sec_code=self.stock_code)
        # return stk_d[self.stock_code]['up_limit']

    @property
    def low_limit(self) -> float:
        # 获取当前交易时间
        # TODO 没有缓存作用，每次访问查询数据库
        trade_date = self.context.current_dt.strftime('%Y-%m-%d')
        df = get_factor(self.stock_code, 'down_limit', trade_date=trade_date, dividend_type=self.context.dividend_type)
        if not df.empty:
            price = df.iloc[0, 0]
        else:
            log.warning(f"无法获取跌停价：sec_code={self.stock_code}, trade_date={trade_date}")
            price = 0.0
        return price

        # # 获取当前交易时间
        # current_dt = pd.Timestamp(self.data.datetime.datetime())
        # trade_date = current_dt.strftime('%Y-%m-%d')
        #
        # # 获取基础行情数据
        # fields = ['up_limit', 'down_limit']
        # # stk_d = crud.get_mkt_stk_d(trade_date=trade_date, fields=fields, stock_lst=[self.stock_code])
        # stk_d = crud.get_mkt_stk_d(trade_date=trade_date, fields=fields, sec_code=self.stock_code)
        # return stk_d[self.stock_code]['down_limit']

    @property
    def paused(self) -> bool:
        # suspend
        # 获取当前交易时间
        # TODO 没有缓存作用，每次访问查询数据库
        trade_date = self.context.current_dt.strftime('%Y-%m-%d')

        # 获取停复牌信息, 默认返回停牌股票的dict，key为股票代码
        suspend_dict = get_suspend_info(trade_date=trade_date, stock_lst=[self.stock_code])
        return True if self.stock_code in suspend_dict else False

    @property
    def is_st(self) -> bool:
        # 获取当前交易时间
        # TODO 没有缓存作用，每次访问查询数据库
        trade_date = self.context.current_dt.strftime('%Y-%m-%d')

        # 获取股票的历史名称
        stock_name = get_stock_name_his(sec_code=self.stock_code, trade_date=trade_date)
        is_st = True if 'ST' in stock_name or '*' in stock_name else False
        # log.debug(f"is_ST: sec_code: {self.stock_code}, trade_date = {trade_date}, name: {stock_name}, ST: {is_st}")
        return is_st

    @property
    def day_open(self) -> float:
        # 获取当前交易时间
        # TODO 没有缓存作用，每次访问查询数据库
        trade_date = self.context.current_dt.strftime('%Y-%m-%d')
        df = get_factor(self.stock_code, 'open', trade_date=trade_date, dividend_type=self.context.dividend_type)
        price = df.iloc[0, 0]
        return price

    @property
    def display_name(self):
        return self.bsc_info_dict[self.stock_code]['fullname']

    @property
    def name(self) -> str:
        # 获取当前交易时间
        trade_date = self.context.current_dt.strftime('%Y-%m-%d')

        # 获取股票的历史名称
        # TODO 没有缓存作用，每次访问查询数据库
        stock_name_dict = get_stock_name_his(sec_code=self.stock_code, trade_date=trade_date)
        return stock_name_dict['name']

    @property
    def industry_name(self) -> str:
        return self.bsc_info_dict[self.stock_code]['industry']

    @property
    def industry_code(self) -> str:
        return self.bsc_info_dict[self.stock_code]['industry']

    @property
    def start_date(self) -> datetime.date:
        # 2007-04-30
        return datetime.datetime.strptime(self.bsc_info_dict[self.stock_code]['list_date'], '%Y-%m-%d').date()

    @property
    def end_date(self):
        end_date = datetime.datetime.strptime('22000101', '%Y%m%d')
        return end_date

    @property
    def type(self):
        # TODO 默认instrument类型为股票
        return 'stock'



import re


def convert_to_type(s: str):
    """
    转换字符串为正确的数据类型

    :param s:
    :return:
    """

    if not isinstance(s, str):
        return s

    s = s.strip()  # 删除两端的空格

    # 检查字符串是否只包含数字
    if s.isdigit():
        return int(s)  # 转换为整数
    elif re.match("^\d+\.\d+$", s):
        return float(s)  # 仅有数字和小数点，则转换为浮点数
    else:
        return s  # 如果既不是纯数字也没有小数点，返回字符串本身


if __name__ == '__main__':
    res = convert_to_type('10')
    print(type(res), res)

    res = convert_to_type('10.3')
    print(type(res), res)

    res = convert_to_type('a10.3')
    print(type(res), res)



# coding=utf-8
from .main import BacktestExecutor
import pqsdk.utils.file_util as fu
import json
import faxdatasdk
from pqsdk import log
import pandas as pd
import quantstats as qs
from pathlib import Path
import datetime
import os


def execute(parameters: dict, script: str = None, strategy_file: str = None):
    # 登录faxdatasdk
    config_file = 'config.sdk.json'
    if not fu.check_path_exists(config_file):
        print("未找到配置文件. 请先运行命令进行配置: pqsdk config")
        exit(-1)

    with open(config_file, 'r', encoding='utf-8') as f:
        sdk_config = json.loads(f.read())
    faxdatasdk.auth_by_token(token=sdk_config['token'], host=sdk_config['host'], audience=sdk_config['audience'])

    if strategy_file:
        strategy_file = os.path.splitext(strategy_file)[0]
    kwargs = {"parameters": parameters, "script": script, "strategy_file": strategy_file}
    executor = BacktestExecutor(kwargs=kwargs)
    results = executor.run()
    return results


def tearsheet(results: dict, save_path="storage/reports"):
    """
    保存回测结果到Tearsheet

    :param save_path: 保存路径
    :param results:
    :return None: .
    """
    log.info("生成Tearsheet, 请稍后...")

    # 策略收益率
    strat_return = results['analysis']['time_return']['strat_return']
    strat_df = pd.DataFrame({'strat_return': strat_return})
    strat_df.index = pd.to_datetime(strat_df.index)
    strat_returns = strat_df['strat_return']

    # benchmark收益率
    bchmk_return = results['analysis']['time_return']['bchmk_return']
    bchmk_df = pd.DataFrame({'bchmk_return': bchmk_return})
    bchmk_df.index = pd.to_datetime(bchmk_df.index)
    # 重命名benchmark的Column名称，用于Tearsheet显示
    bchmk_df = bchmk_df.rename(columns={"bchmk_return": results['benchmark']})
    # bchmk_df = bchmk_df.loc[stats_df.index]  # 以持仓日期列表保留benchmark的记录
    bchmk_returns = bchmk_df[results['benchmark']]

    # tearsheet保存路径
    ts_path = Path(save_path)
    ts_path.mkdir(parents=True, exist_ok=True)
    time_str = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    filename = (
            "tearsheet"
            + "-"
            + time_str.replace("-", "").replace(":", "").replace(" ", "-")
            + ".html"
    )
    filepath = ts_path / filename

    title = f"回测报告 {time_str}"
    qs.reports.html(
        strat_returns,
        benchmark=bchmk_returns,
        title=title,
        output=filepath,
        download_filename=filepath
    )

    log.info(f"Tearsheet创建完成，路径：{filepath}")


def save_metrics(results: dict, save_path="storage/metrics"):
    """
    保存最终回测结果
    :param results:
    :param save_path: 保存路径
    :return:
    """
    # 计算总市值时间序列
    cost = results['analysis']['time_return']['cost']
    market_values = results['analysis']['time_return']['market_value']
    pnls = results['analysis']['time_return']['pnl']
    principal = results['analysis']['time_return']['principal']
    cash = results['analysis']['time_return']['cash']
    asset = results['analysis']['time_return']['asset']
    position_size = results['analysis']['time_return']['position_size']
    nav = results['analysis']['time_return']['nav']
    strat_return = results['analysis']['time_return']['strat_return']
    bchmk_returns = results['analysis']['time_return']['bchmk_return']
    metrics_df = pd.DataFrame({'cost': cost,  # 交易成本
                               'market_value': market_values,  # 总市值
                               'pnl': pnls,  # 盈亏
                               'principal': principal,  # 本金
                               'cash': cash,  # 现金
                               'asset': asset,  # 总资产
                               'position_size': position_size,  # 仓位
                               'nav': nav,  # 净值
                               'strat_return': strat_return,  # 策略收益率
                               'bchmk_return': bchmk_returns  # 基准收益率
                               })
    metrics_df.index = pd.to_datetime(metrics_df.index)
    metrics_df.index.name = "trade_date"
    metrics_df['cum_return'] = (1 + metrics_df['strat_return']).cumprod() - 1  # 累计收益率

    run_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    file_path = Path(save_path)
    file_path.mkdir(parents=True, exist_ok=True)
    filename = ("results" + "_" + run_time.replace("-", "").replace(":", "").replace(" ", "_") + ".csv")
    file_path = file_path / filename
    log.info(f"save metrics data: {file_path}")
    metrics_df.to_csv(file_path, index=True)


def save_orders(results: dict, save_path="storage/orders"):
    orders_df: pd.DataFrame = results['orders']
    run_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    file_path = Path(save_path)
    file_path.mkdir(parents=True, exist_ok=True)
    filename = ("orders" + "_" + run_time.replace("-", "").replace(":", "").replace(" ", "_") + ".csv")
    file_path = file_path / filename
    log.info(f"save orders data: {file_path}")
    orders_df.to_csv(file_path, index=False)


def save_inout_cash(results: dict, save_path="storage/inout_cash"):
    df: pd.DataFrame = results['inout_cash']
    run_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    file_path = Path(save_path)
    file_path.mkdir(parents=True, exist_ok=True)
    filename = ("inout_cash" + "_" + run_time.replace("-", "").replace(":", "").replace(" ", "_") + ".csv")
    file_path = file_path / filename
    log.info(f"save inout_cash data: {file_path}")
    df.to_csv(file_path, index=False)


# ---------------------------------------------------------
# 外部可以访问的列表
# ---------------------------------------------------------
__all__ = ["execute", "tearsheet", "save_metrics", "save_orders", "save_inout_cash"]



import faxdatasdk # noqa
from faxdatasdk import *  # noqa
from .main import *  # noqa
import pqsdk.utils.file_util as fu
import json

# 登录faxdatasdk
config_file = 'config.sdk.json'
if fu.check_path_exists(config_file):
    with open(config_file, 'r', encoding='utf-8') as f:
        sdk_config = json.loads(f.read())
    faxdatasdk.auth_by_token(token=sdk_config['token'], host=sdk_config['host'], audience=sdk_config['audience'])


__all__ = []
__all__.extend(faxdatasdk.__all__)  # noqa
__all__.extend(main.__all__)  # noqa



from six import with_metaclass
import abc


class AbstractRunInfo(with_metaclass(abc.ABCMeta)):
    """
    策略运行信息，包括回测过程中的所有参数。
    """

    @property
    @abc.abstractmethod
    def tenant_id(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def portfolio_id(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def parameters(self):
        """
        策略中自定义参数字典
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def start_date(self):
        """
        策略的开始日期
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def end_date(self):
        """
        策略的结束日期
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def stock_starting_cash(self):
        """
        股票账户初始资金
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def strategy_id(self):
        """
        策略id
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def release_id(self):
        """
        策略版本id
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def strategy_name(self):
        """
        策略名称
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def strategy_remark(self):
        """
        策略备注
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def is_queued_order(self):
        """
        策略是支持委托队列，如果False，则直接通过交易终端进行交易
        如果是回测，直接返回False，直接交易。
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def benchmark(self):
        """
        基准指数代码，比如沪深300
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def stock_pool(self) -> list:
        """
        股票池，指数列表
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def stock_pool_members(self) -> list:
        """
        当日的股票池成员
        :return:
        """
        raise NotImplementedError
